SuperStrict

Rem
bbdoc: Holzchopf/MaxGuiCalendar
about: Calendar Proxygadget for MaxGUI, written for miniBCC #11
End Rem
'Module Holzchopf.MaxGuiCalendar

'ModuleInfo "Version: 1.00"
'ModuleInfo "Author: Holzchopf"
'ModuleInfo "License: Public Domain"

Import MaxGUI.Drivers
Import BRL.EventQueue
Import Pub.StdC

' Calendar event
Rem
bbdoc: EVENT_DATESELECTED is emitted when a date is selected
about: #EventData contains selected date as integer.
Only works between 1901 and 2038.
If you allow a longer period, get the timestamp (long) by calling the @GetSelectedDate Method.
End Rem
Const EVENT_DATESELECTED:Int	= $2100
TEvent.RegisterId EVENT_DATESELECTED, "DateSelected"

Rem
bbdoc: Calendar gadget type
End Rem
Type TCalendarGadget Extends TProxyGadget
	Rem
	bbdoc: minimum selectable year
	End Rem
	Field MIN_YEAR:Int
	Field min_date:Int
	Rem
	bbdoc: maximum selectable year
	End Rem
	Field MAX_YEAR:Int
	Field max_date:Int
	
	Rem
	bbdoc: menu id for month menu (also uses the next 11 values)
	End Rem
	Global MENU_CAL_MONTH:Int	= $01010000
	Rem
	bbdoc: menu id for year enter menu
	End Rem
	Global MENU_CAL_YENTER:Int	= $01011000
	Rem
	bbdoc: menu id for year menu (also uses the next 9999 values)
	End Rem
	Global MENU_CAL_YEAR:Int		= $01012000
	
	Rem
	bbdoc: background-color, red
	End Rem
	Field COLOR_BG_R:Byte		= 255
	Rem
	bbdoc: background-color, green
	End Rem
	Field COLOR_BG_G:Byte		= 255
	Rem
	bbdoc: background-color, blue
	End Rem
	Field COLOR_BG_B:Byte		= 255
	Rem
	bbdoc: background-color, highlighted, red
	End Rem
	Field COLOR_HL_R:Byte		= 224
	Rem
	bbdoc: background-color, highlighted, green
	End Rem
	Field COLOR_HL_G:Byte		= 240
	Rem
	bbdoc: background-color, highlighted, blue
	End Rem
	Field COLOR_HL_B:Byte		= 255
	Rem
	bbdoc: font-color, red
	End Rem
	Field COLOR_FONT_R:Byte		= 0
	Rem
	bbdoc: font-color, green
	End Rem
	Field COLOR_FONT_G:Byte		= 0
	Rem
	bbdoc: font-color, blue
	End Rem
	Field COLOR_FONT_B:Byte		= 0
	Rem
	bbdoc: font-color of days not in month, red
	End Rem
	Field COLOR_EXDAYS_R:Byte	= 192
	Rem
	bbdoc: font-color of days not in month, green
	End Rem
	Field COLOR_EXDAYS_G:Byte	= 192
	Rem
	bbdoc: font-color of days not in month, blue
	End Rem
	Field COLOR_EXDAYS_B:Byte	= 192
	Rem
	bbdoc: font-color of today, red
	End Rem
	Field COLOR_TODAY_R:Byte		= 0
	Rem
	bbdoc: font-color of today, green
	End Rem
	Field COLOR_TODAY_G:Byte		= 128
	Rem
	bbdoc: font-color of today, blue
	End Rem
	Field COLOR_TODAY_B:Byte		= 255
	Rem
	bbdoc: background-color of marked day, red
	End Rem
	Field COLOR_MARKED_R:Byte	= 0
	Rem
	bbdoc: background-color of marked day, green
	End Rem
	Field COLOR_MARKED_G:Byte	= 0
	Rem
	bbdoc: background-color of marked day, blue
	End Rem
	Field COLOR_MARKED_B:Byte	= 255
	
	Field gdWin:TGadget
	Field gdMain:TGadget
	Field gdChilds:TGadget[][,] = New TGadget[][8,7]
	Field gdMonthBack:TGadget
	Field gdMonth:TGadget
	Field gdMonthFore:TGadget
	Field gdMonthMenu:TGadget
	Field gdMonthMenuItem:TGadget[]
	Field gdYearBack:TGadget
	Field gdYear:TGadget
	Field gdYearMenu:TGadget
	Field gdYearTF:TGadget
	Field gdYearFore:TGadget
	Field evHotkey:THotKey
	Field fntBold:TGuiFont
	Field fntNorm:TGuiFont

	Field mark:Int
	Field navdate:Int
	Field dates:Int[][,] = New Int[][8,7]
	Field dateSel:Int[]
	Rem
	bbdoc: caption for calendar week, default: KW
	End Rem
	Field weekCaption:String="KW"
	Rem
	bbdoc: caption for days
	End Rem
	Field days:String[]=["Mo","Di","Mi","Do","Fr","Sa","So"]
	Rem
	bbdoc: short forms of months
	End Rem
	Field monthsShort:String[]=["Jan","Feb","Mrz","Apr","Mai","Jun","Jul","Aug","Sep","Okt","Nov","Dez"]
	Rem
	bbdoc: full names of months
	End Rem
	Field months:String[]=["Januar","Februar","Mrz","April","Mai","Juni","Juli","August","September","Oktober","November","Dezember"]
	
	Rem
	bbdoc: Create a calendar gadget
	End Rem
	Function Create:TCalendarGadget(pX:Int, pY:Int, pW:Int, pH:Int, pGroup:TGadget=Null)
		Local c:TCalendarGadget = New TCalendarGadget
		' load standard and bold font variant
		c.fntBold = LookupGuiFont(GUIFONT_SYSTEM, , FONT_BOLD)
		c.fntNorm = LookupGuiFont(GUIFONT_SYSTEM)
		' find parent window gadget
		c.gdWin = pGroup
		While c.gdWin.parent And c.gdWin.Class()<>GADGET_WINDOW
			c.gdWin = c.gdWin.parent
		Wend
		' set colors
		LookupGuiColor(GUICOLOR_GADGETBG, c.COLOR_BG_R, c.COLOR_BG_G, c.COLOR_BG_B)
		LookupGuiColor(GUICOLOR_GADGETFG, c.COLOR_FONT_R, c.COLOR_FONT_G, c.COLOR_FONT_B)
		c.COLOR_EXDAYS_R = c.COLOR_BG_R *3 /4 +c.COLOR_FONT_R /4
		c.COLOR_EXDAYS_G = c.COLOR_BG_G *3 /4 +c.COLOR_FONT_G /4
		c.COLOR_EXDAYS_B = c.COLOR_BG_B *3 /4 +c.COLOR_FONT_B /4
		LookupGuiColor(GUICOLOR_LINKFG, c.COLOR_TODAY_R, c.COLOR_TODAY_G, c.COLOR_TODAY_B)
		LookupGuiColor(GUICOLOR_SELECTIONBG, c.COLOR_MARKED_R, c.COLOR_MARKED_G, c.COLOR_MARKED_B)
		' set min/max date initially
		c.SetMinMaxYear(1200,3200)
		' main gadget that contains all calendar gadgets
		c.gdMain = CreatePanel(pX,pY,pW,pH,pGroup)
		SetGadgetLayout(c.gdMain, EDGE_ALIGNED, EDGE_ALIGNED, EDGE_ALIGNED, EDGE_ALIGNED)
		For Local v:Int = 0 Until 8
			Local y:Int = pH / 8.0 * v
			Local h:Int = (pH / 8.0 * (v+1)-y)
			For Local u:Int = 0 Until 8
				Local x:Int = pW / 8.0 * u
				Local w:Int = (pW / 8.0 * (u+1)-x)
				' captions and days
				If v<7
					Local panel:TGadget = CreatePanel(x,y,w,h, c.gdMain)
					SetGadgetLayout(panel, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					If u>0
						SetGadgetColor(panel, c.COLOR_BG_R, c.COLOR_BG_G, c.COLOR_BG_B)
					EndIf
					Local label:TGadget = CreateLabel("",0,2,w,h-2, panel, LABEL_CENTER)
					SetGadgetLayout(label, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					' sensitize days for mouse-input
					If v>0 And u>0
						SetGadgetSensitivity(label, SENSITIZE_MOUSE)
					EndIf
					' captions; set bold font
					If v=0
						If u=0
							SetGadgetText(label,c.weekCaption)
						Else
							SetGadgetText(label,c.days[u-1])
						EndIf
						SetGadgetFont(label, c.fntBold)
					EndIf
					' save gadgets
					c.gdChilds[u,v] = [panel,label]
				' buttons for navigation
				Else
					' month navigation
					If u=0
						c.gdMonthBack = CreateButton(Chr(9664),x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdMonthBack, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					ElseIf u=1
						w = pW/8.0*(u+2)-x
						c.gdMonth = CreateButton("",x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdMonth, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
						c.gdMonthMenu = CreateMenu("",0, c.gdMain)
						Local _m:Int=0
						For Local m:String = EachIn c.months
							c.gdMonthMenuItem :+ [CreateMenu(m, MENU_CAL_MONTH + _m, c.gdMonthMenu)]
							_m:+1
						Next
					ElseIf u=3
						c.gdMonthFore = CreateButton(Chr(9654),x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdMonthFore, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					' year navigation
					ElseIf u=4
						c.gdYearBack = CreateButton(Chr(9664),x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdYearBack, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					ElseIf u=5
						w = pW/8.0*(u+2)-x
						c.gdYear = CreateButton("",x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdYear, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
						c.gdYearTF = CreateTextField(x,y,w,h, c.gdMain)
						HideGadget(c.gdYearTF)
					ElseIf u=7
						c.gdYearFore = CreateButton(Chr(9654),x,y,w,h, c.gdMain)
						SetGadgetLayout(c.gdYearFore, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE)
					EndIf
				EndIf
			Next
		Next
		' initially, mark current date
		c.MarkCurrentDate()
		Return c
	End Function
	
	Rem
	bbdoc: set minimum / maximum year
	End Rem
	Method SetMinMaxYear(pMin:Int, pMax:Int)
		MIN_YEAR = pMin
		MAX_YEAR = pMax
		min_date = _DMYToDatestamp(1,1, MIN_YEAR)
		max_date = _DMYToDatestamp(31,12, MAX_YEAR)
	End Method
	
	Rem
	bbdoc: get timestamp of current date
	End Rem
	Function GetCurrentDate:Long()
		Local time:Int
		time_(Varptr time)
		Return time
	End Function
	
	Rem
	bbdoc: mark current date
	End Rem
	Method MarkCurrentDate()
		mark = _TimeToDatestamp( GetCurrentDate() )
		navdate = mark
		Local date:Int[] = _DatestampToDMY(mark)
		SetGadgetText(gdMonth, monthsShort[date[1]-1])
		SetGadgetText(gdYear, date[2])
		_Build()
	End Method
	Rem
	bbdoc: mark specific date
	End Rem
	Method MarkDate(pTimestamp:Long)
		mark = _TimeToDatestamp(pTimestamp)
		_Build()
	End Method
	Rem
	bbdoc: navigate to marked date
	End Rem
	Method NavigateToMarked()
		_Navigate(mark)
	End Method
	Rem
	bbdoc: navigate to specific date
	End Rem
	Method NavigateToDate(pTimestamp:Long)
		_Navigate( _TimeToDatestamp(pTimestamp) )
	End Method
	Method _Navigate(pDatestamp:Int)
		navdate = Min(max_date, Max(min_date, pDatestamp))
	End Method
	
	Rem
	bbdoc: returns timestamp (long) of selected date
	End Rem
	Method GetSelectedDate:Long()
		Return _DMYToTimestamp(dateSel[0], dateSel[1], dateSel[2])
	End Method

	' timestamp to [day:int, month:int, year:int]
	Function _TimestampToDMY:Int[](pTimestamp:Long)
		Return _DatestampToDMY( _TimeToDatestamp(pTimestamp) )
	End Function
	' datestamp to [day:int, month:int, year:int]
	Function _DatestampToDMY:Int[](pDatestamp:Int)
		Local year:Int, month:Int, day:Int
		' determine year
		Local days:Int = pDatestamp
		' count years
		Local cy400:Int = days / 146097		' 146097 days in 400 years (36524 x 4 + 1)
		days :- cy400 * 146097
		Local cy100:Int = days / 36524		' 36524 days in 100 years (1461 x 5 - 1)
		days :- cy100 * 36524
		Local cy4:Int = days / 1461			' 1461 days in 4 years (365 x 4 + 1)
		days :- cy4 * 1461
		Local cy1:Int = days / 365			' 365 days in one year
		' restore year
		year = cy400*400 + cy100*100 + cy4*4 + cy1
		' leap years
		Local lyears:Int = (year-1) / 400 - (year-1) / 100 + (year-1) / 4
		' day in year
		Local yday:Int = pDatestamp -lyears -year*365 +1
		' in non-leap years, add one day after 28. feb for further calculations
		If Not _IsLeapYear(year) And yday>59 Then yday:+1
		If yday<=31
			month=1
			day=yday
		ElseIf yday<=60
			month=2
			day=yday-31
		ElseIf yday<=91
			month=3
			day=yday-60
		ElseIf yday<=121
			month=4
			day=yday-91
		ElseIf yday<=152
			month=5
			day=yday-121
		ElseIf yday<=182
			month=6
			day=yday-152
		ElseIf yday<=213
			month=7
			day=yday-182
		ElseIf yday<=244
			month=8
			day=yday-213
		ElseIf yday<=274
			month=9
			day=yday-244
		ElseIf yday<=305
			month=10
			day=yday-274
		ElseIf yday<=335
			month=11
			day=yday-305
		Else
			month=12
			day=yday-335
		EndIf
		Return [day,month,year]
	End Function
	' date to timestamp
	Function _DMYToTimestamp:Long(pDay:Int, pMonth:Int, pYear:Int)
		Return _DateToTimestamp( _DMYToDatestamp(pDay, pMonth, pYear) )
	End Function
	' date to datestamp
	Function _DMYToDatestamp:Long(pDay:Int, pMonth:Int, pYear:Int)
		' days in all those years (leap years already included)
		Local dyear:Int = pYear*365 +(pYear-1)/4 -(pYear-1)/100 +(pYear-1)/400
		' day in that year
		Local yday:Int = _GetYearDay(pDay, pMonth, pYear)
		' total days since 01.01.0000
		Local datestamp:Int = dyear +yday
		Return datestamp
	End Function
	' returns day in year
	Function _GetYearDay:Int(pDay:Int, pMonth:Int, pYear:Int)
		Local yday:Int = [0,31,60,91,121,152,182,213,244,274,305,335][pMonth-1]
		' in non-leap years, substract one day after 28. feb
		If Not _IsLeapYear(pYear) And yday>59 Then yday:-1
		yday :+ pDay -1
		Return yday
	End Function
	' returns true if year is leap year
	Function _IsLeapYear:Int(pYear:Int)
		If (pYear Mod 400)=0 Then Return True
		If (pYear Mod 4)=0 And (pYear Mod 100)<>0 Then Return True
		Return False
	End Function
	' returns day of week, 0 = monday
	Function _GetWeekDay:Int(pDay:Int,pMonth:Int,pYear:Int)
		Local datestamp:Int = _DMYToDateStamp(pDay, pMonth, pYear)
		Return (datestamp+6) Mod 7
	End Function
	' returns week in year, beginning at the week with 4.Jan
	Function _GetWeek:Int(pDay:Int,pMonth:Int,pYear:Int)
		Local datestamp:Int = _DMYToDatestamp(pDay, pMonth, pYear)
		Local dsWeekThu:Long = datestamp +3 -_GetWeekDay(pDay, pMonth, pYear)
		Local dateWeekThu:Int[] = _DatestampToDMY(dsWeekThu)
		Local yday:Int = _GetYearDay(dateWeekThu[0], dateWeekThu[1], dateWeekThu[2])
		Local dFirstThu:Int = (10 - _GetWeekDay(1, 1, pYear)) Mod 7
		Return (yday - dFirstThu) / 7 +1
	End Function
	' convert from seconds since 1.1.1970 to days since 1.1.0000
	Function _TimeToDatestamp:Int(pTimestamp:Long)
		Local datestamp:Int
		datestamp = pTimestamp / 86400				' it's now days since 01.01.1970
		datestamp :+ 4*146097 +3*36524 +17*1461 +2*365	' and now days since 01.01.0000
		Return datestamp
	End Function
	' convert from days since 1.1.0000 to seconds since 1.1.1970
	Function _DateToTimestamp:Long(pDatestamp:Int)
		Local timestamp:Long
		timestamp = pDatestamp -(4*146097 +3*36524 +17*1461 +2*365)	' now days since 01.01.1970
		timestamp :* 86400									' now seconds since 01.01.1970
		Return timestamp
	End Function
	
	' build the gadget
	Method _Build()
		' find day to start with
		Local today:Int = _TimeToDatestamp(GetCurrentDate())
		Local date:Int[] = _DatestampToDMY(navdate)
		Local month:Int = date[1]
		Local year:Int = date[2]
		Local wday:Int = _GetWeekDay(1, date[1], date[2])
		Local datestamp:Int = _DMYToDatestamp(1, date[1], date[2]) -wday
		' go through table
		For Local v:Int = 1 Until 7
			For Local u:Int = 1 Until 8
				date = _DatestampToDMY(datestamp)
				dates[u,v] = date
				' only label available dates
				If date[2] >= MIN_YEAR And date[2] <= MAX_YEAR
					SetGadgetText(gdChilds[u,v][1], date[0])
					' color days in and not in month
					If date[1]=month
						SetGadgetTextColor(gdChilds[u,v][1], COLOR_FONT_R, COLOR_FONT_R, COLOR_FONT_R)
					Else
						SetGadgetTextColor(gdChilds[u,v][1], COLOR_EXDAYS_R, COLOR_EXDAYS_R, COLOR_EXDAYS_R)
					EndIf
					' highlight current date
					If datestamp = today
						SetGadgetFont(gdChilds[u,v][1], fntBold)
						SetGadgetTextColor(gdChilds[u,v][1], COLOR_TODAY_R, COLOR_TODAY_G, COLOR_TODAY_B)
					Else
						SetGadgetFont(gdChilds[u,v][1], fntNorm)
					EndIf
					' highlight marked date
					_ColorBG(u,v)
				' remove labels of days outside range
				Else
					SetGadgetText(gdChilds[u,v][1], "")
				EndIf
				datestamp :+ 1
				' enumerate weeks
				If u=1
					Local week:Int = _GetWeek(date[0],date[1],date[2])
					SetGadgetText(gdChilds[0,v][1], week)
				EndIf
			Next
		Next
		' update month/year buttons
		SetGadgetText(gdMonth, monthsShort[month-1])
		SetGadgetText(gdYear, year)
	End Method
	
	Rem
	bbdoc: update calendar gadget
	End Rem
	Method Update()
		' set / remove hotkey for gdYearTF
		If ActiveGadget() = gdYearTF
			evHotkey = SetHotKeyEvent(KEY_ENTER,0,CreateEvent(EVENT_HOTKEYHIT, gdYearTF))
		Else
			RemoveHotKey(evHotkey)
		EndIf
		' select event
		Select EventSource()
		' previous month
		Case gdMonthBack
			Local date:Int[] = _DatestampToDMY(navdate)
			date[1] :- 1
			If date[1]=0
				date[1] = 12
				date[2] :- 1
			EndIf
			_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
			_Build()
		' next month
		Case gdMonthFore
			Local date:Int[] = _DatestampToDMY(navdate)
			date[1] :+ 1
			If date[1]=13
				date[1] = 0
				date[2] :+ 1
			EndIf
			_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
			_Build()
		' select month
		Case gdMonth
			Local date:Int[] = _DatestampToDMY(navdate)
			' loop prevents bug of double-selections when exiting menu by clicking on month-switch-button
			For Local m:Int = 0 Until 12
				UncheckMenu(gdMonthMenuItem[m])
			Next
			CheckMenu(gdMonthMenuItem[date[1]-1])
			PopupWindowMenu(gdWin,gdMonthMenu)
		' previous year
		Case gdYearBack
			Local date:Int[] = _DatestampToDMY(navdate)
			date[2] :- 1
			_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
			_Build()
		' next year
		Case gdYearFore
			Local date:Int[] = _DatestampToDMY(navdate)
			date[2] :+ 1
			_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
			_Build()
		' select year
		Case gdYear
			Local date:Int[] = _DatestampToDMY(navdate)
			' create menu starting 5 yrs before current year ending 5yrs after
			gdYearMenu = CreateMenu("",0,gdMain)
			CreateMenu("...", MENU_CAL_YENTER, gdYearMenu)
			CreateMenu("",0, gdYearMenu)
			For Local _y:Int = -5 To 5
				Local y:Int = date[2] +_y
				If y >= MIN_YEAR And y <= MAX_YEAR
					Local gdItem:TGadget = CreateMenu(y, MENU_CAL_YEAR+y, gdYearMenu)
					If _y=0
						CheckMenu(gdItem)
					EndIf
				EndIf
			Next
			PopupWindowMenu(gdWin, gdYearMenu)
		' enter year
		Case gdYearTF
			Select EventID()
			' when gadget lost focus or KEY_ENTER was hit
			Case EVENT_GADGETLOSTFOCUS, EVENT_HOTKEYHIT
				Local date:Int[] = _DatestampToDMY(navdate)
				date[2] = Int(GadgetText(gdYearTF))
				_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
				_Build()
				HideGadget(gdYearTF)
				ShowGadget(gdYear)
			End Select
		' other events
		Default
			Select EventID()
			' month/year menu
			Case EVENT_MENUACTION
				Local data:Int = EventData()
				' enter year
				If data = MENU_CAL_YENTER
					Local date:Int[] = _DatestampToDMY(navdate)
					SetGadgetText(gdYearTF, date[2])
					HideGadget(gdYear)
					ShowGadget(gdYearTF)
					ActivateGadget(gdYearTF)
				' select month
				ElseIf data >= MENU_CAL_MONTH And data < MENU_CAL_MONTH+12
					Local date:Int[] = _DatestampToDMY(navdate)
					date[1] = data -MENU_CAL_MONTH +1
					_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
					_Build()
				' select year
				ElseIf data >= MENU_CAL_YEAR And data <= MENU_CAL_YEAR+9999
					Local date:Int[] = _DatestampToDMY(navdate)
					date[2] = data -MENU_CAL_YEAR
					_Navigate( _DMYToDatestamp(date[0], date[1], date[2]) )
					_Build()
				EndIf
			' highlight days
			Case EVENT_MOUSEENTER
				Local u:Int, v:Int
				If _GetEventSourceUV(u,v)
					If dates[u,v][2] => MIN_YEAR And dates[u,v][2] <= MAX_YEAR
						SetPointer(POINTER_HAND)
						_ColorBG(u,v, True)
					EndIf
				EndIf
			Case EVENT_MOUSELEAVE, EVENT_GADGETLOSTFOCUS
				Local u:Int, v:Int
				If _GetEventSourceUV(u,v)
					SetPointer(POINTER_DEFAULT)
					_ColorBG(u,v)
				EndIf
			' select date
			Case EVENT_MOUSEDOWN
				Local u:Int, v:Int
				If _GetEventSourceUV(u,v)
					dateSel = dates[u,v]
				EndIf
			Case EVENT_MOUSEUP
				Local u:Int, v:Int
				If _GetEventSourceUV(u,v)
					' only if selected date is in range
					If dates[u,v][2] => MIN_YEAR And dates[u,v][2] <= MAX_YEAR
						' only if MouseUp() happens in MouseDown()-gadget
						If EventX()>=0 And EventY()<GadgetWidth(gdChilds[u,v][1])
							If EventY()>=0 And EventY()<GadgetHeight(gdChilds[u,v][1])
								' emit calendar event
								Local str:String = DateToString(dateSel[0], dateSel[1], dateSel[2])
								Local date:Long = _DMYToTimestamp(dateSel[0], dateSel[1], dateSel[2])
								EmitEvent(CreateEvent(EVENT_DATESELECTED, gdMain, date, , , , str))
								MarkDate(date)
								_Build()
							EndIf
						EndIf
					EndIf
				EndIf
			End Select
		End Select
	End Method

	Method _ColorBG(pU:Int, pV:Int, pHL:Int=False)
		Local datestamp:Int = _DMYToDatestamp(dates[pU,pV][0], dates[pU,pV][1], dates[pU,pV][2])
		If datestamp = mark
			SetGadgetColor(gdChilds[pU,pV][0], COLOR_MARKED_R, COLOR_MARKED_G, COLOR_MARKED_B)
			SetGadgetTextColor(gdChilds[pU,pV][1], COLOR_BG_R, COLOR_BG_G, COLOR_BG_B)
		ElseIf pHL
			SetGadgetColor(gdChilds[pU,pV][0], COLOR_HL_R, COLOR_HL_G, COLOR_HL_B)
		Else
			SetGadgetColor(gdChilds[pU,pV][0], COLOR_BG_R, COLOR_BG_G, COLOR_BG_B)
		EndIf
	End Method

	' returns u/v coordinate of gadget that triggered the event
	Method _GetEventSourceUV:Int(pU:Int Var, pV:Int Var)
		For Local v:Int = 1 Until 7
			For Local u:Int = 1 Until 8
				If EventSource() = gdChilds[u,v][1]
					pU = u
					pV = v
					Return True
				EndIf
			Next
		Next
		Return False
	End Method
	
	Rem
	bbdoc: create string formatted dd.mm.yyyy from day, month, year
	End Rem
	Function DateToString:String(pDay:Int, pMonth:Int, pYear:Int)
		Local strD:String = "0"+pDay
		Local strM:String = "0"+pMonth
		Local strY:String = "000"+pYear
		Return strD[strD.length-2..]+"."+strM[strM.length-2..]+"."+strY[strY.length-4..]
	End Function
End Type

Rem
bbdoc: create calendar gadget
End Rem
Function CreateCalendarGadget:TCalendarGadget(pX:Int, pY:Int, pW:Int, pH:Int, pGroup:TGadget=Null)
	Return TCalendarGadget.Create(pX,pY,pW,pH,pGroup)
End Function

Rem
bbdoc: update calendar gadget
End Rem
Function UpdateCalendarGadget(pCalendar:TGadget)
	TCalendarGadget(pCalendar).Update()
End Function